/**
 *
 */
package org.dmatej.util;

import java.util.LinkedList;

import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.spi.LoggingEvent;

/**
 * LOG4J Appender - catches log events into a FIFO cache. Logs are later returned and removed via the
 * {@link #pop()} method and can be for example checked. Cache can contain up to
 * {@value #MAX_LOGS_DEFAULT} events at the same time.
 * <p>
 * <b>Usage:</b><br>
 * 1) Setup this appender for some logger:<br>
 * <code>
 * log4j.appender.TEST_LOG_COLLECTOR=cz.i.cis.testutils.log.EventCollectorAppender
 * log4j.logger.SomeLogger=INFO, TEST_LOG_COLLECTOR
 * </code>
 * <p>
 * 2) Access it in junit test:<br>
 * <code>
 * CollectorAppender appender = (EventCollectorAppender) AUDITLOG.getAppender("TEST_LOG_COLLECTOR");
 * </code>
 * <p>
 * 3) Compare results:<br>
 * <code>
 * assertEquals(..., appender.pop()...);
 * </code>
 * <p>
 * 4) After each junit test drop the collected events (if some remained):</br> <code>
 * appender.clearCache();
 * </code>
 *
 * @author David Matějček
 */
public class EventCollectorAppender extends AppenderSkeleton {

  private static final int MAX_LOGS_DEFAULT = 8192;

  private int capacity;
  private LinkedList<LoggingEvent> logs = new LinkedList<LoggingEvent>();


  /**
   * Creates new initialized instance.
   */
  public EventCollectorAppender() {
    super();
    this.capacity = MAX_LOGS_DEFAULT;
  }


  /**
   * Set the maximal count of logs that could be hold. If the capacity is reached, further logs will
   * be dropped.
   *
   * @param capacity
   */
  public void setCapacity(final int capacity) {
    this.capacity = capacity;
  }


  /**
   * @return the maximal count of logs that could be hold.
   */
  public int getCapacity() {
    return this.capacity;
  }


  /**
   * Returns and deletes the first log message from the FIFO cache.
   *
   * @return the first log event from the cache. Null if empty.
   */
  public synchronized LoggingEvent pop() {
    return this.logs.pollFirst();
  }


  /**
   * Cleares all messages from the FIFO cache
   */
  public synchronized void clearCache() {
    this.logs.clear();
  }


  /**
   * @return count of the logs in cache.
   */
  public synchronized int size() {
    return this.logs.size();
  }


  @Override
  public synchronized void close() {
    this.logs = new LinkedList<LoggingEvent>();
  }


  @Override
  public synchronized boolean requiresLayout() {
    return false;
  }


  /**
   * If collector cache contains less than #MAX_LOGS_DEFAULT records, puts the logMessage into
   * cache. Otherwise writes a warning to System.err.
   */
  @Override
  protected void append(final LoggingEvent event) {
    if (logs.size() < MAX_LOGS_DEFAULT) {
      logs.add(event);
    } else {
      System.err.println("WARN: Log collector overflow. Holding " + size() + " logs in memory.");
    }

  }
}
