/**
 *
 */
package org.glassfish;

import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.URL;
import java.util.concurrent.TimeUnit;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.dmatej.util.EventCollectorAppender;
import org.glassfish.embeddable.CommandResult;
import org.glassfish.embeddable.CommandRunner;
import org.glassfish.embeddable.GlassFishException;
import org.jboss.arquillian.container.test.api.Deployment;
import org.jboss.arquillian.container.test.api.Testable;
import org.jboss.arquillian.junit.Arquillian;
import org.jboss.shrinkwrap.api.ShrinkWrap;
import org.jboss.shrinkwrap.api.spec.EnterpriseArchive;
import org.jboss.shrinkwrap.api.spec.JavaArchive;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.rules.TestName;
import org.junit.runner.RunWith;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.htmlunit.HtmlUnitDriver;
import org.openqa.selenium.support.events.EventFiringWebDriver;
import org.openqa.selenium.support.events.WebDriverEventListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.bridge.SLF4JBridgeHandler;

import com.sun.enterprise.security.auth.realm.file.FileRealm;

/**
 * Arquillian test infrastructure; creating glassfish instance, generating of EAR file and deploy.
 *
 * @author David Matějček
 */
@RunWith(Arquillian.class)
public abstract class ArquillianTest {

  private static final Logger LOG = LoggerFactory.getLogger(ArquillianTest.class);

  // see arquillian.xml
  protected final String url = "http://localhost:10082";

  private static EnterpriseArchive ear;
  protected EventCollectorAppender errorAppender;
  protected WebDriver driver;

  @Rule
  public TestName name = new TestName();
  private static boolean alreadyInitialized;


  /**
   * Initialize test class.
   *
   * @return {@link JavaArchive} to deploy to the container.
   * @throws Exception exception
   */
  @Deployment
  public static EnterpriseArchive getEar() throws Exception {
    if (alreadyInitialized) {
      LOG.trace("Already initialized, returning.");
      return ear;
    }
    // do not try it again - if it fails, it should fail again.
    alreadyInitialized = true;
    initEnvironment();

    final JavaArchive ejbs = ShrinkWrap.create(JavaArchive.class, "ejbs.jar");
    ejbs.addClasses(SomeEjb.class).addAsManifestResource("common/beans.xml", "beans.xml");
    LOG.info("{}", ejbs.toString(true));

    final JavaArchive war1 = ShrinkWrap.create(JavaArchive.class, "war1.war").addAsResource("webapp1", "");
    // UNCOMMENT THIS LINE AND THE TEST WILL NOT FAIL!
//    war1.addAsManifestResource("common/beans.xml", "beans.xml");
    war1.addClasses(SomeServlet.class);
    LOG.info("{}", war1.toString(true));

    // Testable - without that Arquillian reports that it doesn't know which archive to test.
    final JavaArchive war2 = ShrinkWrap.create(JavaArchive.class, "war2.war");
    war2.addAsResource("webapp2", "");
    LOG.info("{}", war2.toString(true));

    ear = ShrinkWrap.create(EnterpriseArchive.class, "ear.ear").addAsManifestResource("application.xml")
        .addAsModule(war1).addAsModule(ejbs).addAsModule(Testable.archiveToTest(war2));
    LOG.info("ear:\n{}", ear.toString(true));
    return ear;
  }

  @Before
  public void startSelenium() {
    LOG.trace("startSelenium()");

    SLF4JBridgeHandler.removeHandlersForRootLogger();
    SLF4JBridgeHandler.install();

    final HtmlUnitDriver currentDriver = new HtmlUnitDriver();
    currentDriver.setJavascriptEnabled(true);
    driver = new EventFiringWebDriver(currentDriver);
    driver.manage().timeouts().implicitlyWait(5, TimeUnit.SECONDS);
    driver.manage().timeouts().setScriptTimeout(10, TimeUnit.SECONDS);

    final org.apache.log4j.Logger logger = org.apache.log4j.Logger.getLogger("javax.enterprise.web.core");
    errorAppender = (EventCollectorAppender) logger.getAppender("TEST_COLLECTOR");
    errorAppender.setCapacity(1);
  }


  @After
  public void stopSelenium() throws Exception {

    LOG.debug("stopSelenium()");
    driver.quit();
    driver = null;
  }

  private static void initEnvironment() {
    LOG.debug("initEnvironment()");
    LOG.debug("System properties:\n  {}", System.getProperties());

    try {
      runCommand("set", "configs.config.server-config.jms-service.type=EMBEDDED");
      runCommand("set", "configs.config.server-config.jms-service.jms-host.default_JMS_host.port=" + getFreePort());
      runCommand("set", "configs.config.server-config.iiop-service.iiop-listener.orb-listener-1.port=" + getFreePort());
      runCommand("set", "configs.config.server-config.iiop-service.iiop-listener.SSL.port=" + getFreePort());
      runCommand("set", "configs.config.server-config.iiop-service.iiop-listener.SSL_MUTUALAUTH.port=" + getFreePort());
      runCommand("set", "configs.config.server-config.http-service.virtual-server.server.property.sso-enabled=true");
      runCommand("set", "configs.config.server-config.http-service.virtual-server.server.sso-enabled=true");

      final URL loginfile = ArquillianTest.class.getResource("/login.conf");
      System.setProperty("java.security.auth.login.config", new File(loginfile.toURI()).getAbsolutePath());

      final URL passwordfile = ArquillianTest.class.getResource("/passwordfile.txt");
      // FIXME: another bug in glassfish: org.jvnet.hk2.config.UnsatisfiedDependencyException:
      // injection failed on com.sun.enterprise.security.cli.CreateFileUser.userpassword with class
      // java.lang.String
      // runCommand("create-file-user", "--passwordfile", new File(passwordfile.toURI()).getAbsolutePath(), "--authrealmname", "file", "tester");
      runCommand("create-auth-realm", "--classname", FileRealm.class.getCanonicalName(), "--property", "file="
          + new File(passwordfile.toURI()).getAbsolutePath() + ":jaas-context=fileRealm", "fileRealm");

    } catch (final Exception e) {
      throw new IllegalStateException(e);
    }
  }


  /**
   * Execute the command with parameters and return a result.
   *
   * @param command
   * @param parameters
   * @return result of the command
   * @throws GlassFishException - cannot communicate with the instance
   * @throws IllegalStateException - invalid parameters or command
   */
  private static CommandResult runCommand(final String command, final String... parameters) throws GlassFishException {
    LOG.debug("runCommand(command={}, parameters={})", command, parameters);
    try {
      final InitialContext ctx = new InitialContext();
      final CommandRunner runner = (CommandRunner) ctx.lookup("org.glassfish.embeddable.CommandRunner");
      if (runner == null) {
        throw new IllegalStateException("No command runner instance found in initial context!");
      }

      final CommandResult result = runner.run(command, parameters);
      checkCommandResult(command, result);
      return result;
    } catch (final NamingException e) {
      throw new IllegalStateException("Cannot run command " + command, e);
    }
  }


  private static void checkCommandResult(final String cmd, final CommandResult result) {
    LOG.info("Command: {}\n  Result.status:\n  {}\n  Result.out:\n  {}\n  Result.failCause:\n  {}\n", new Object[] {
        cmd, result.getExitStatus(), result.getOutput(), result.getFailureCause()});

    if (result.getExitStatus().ordinal() != 0) {
      throw new IllegalStateException("Command '" + cmd + "' was unsuccessful: " + result.getOutput(),
          result.getFailureCause());
    }
  }


  /**
   * Tries to alocate a free local port.
   *
   * @return a free local port number.
   * @throws IllegalStateException if it fails for 20 times
   */
  private static int getFreePort() throws IllegalStateException {
    int counter = 0;
    while (true) {
      counter++;
      try {
        final ServerSocket socket = new ServerSocket(0);
        final int port = socket.getLocalPort();
        socket.setSoTimeout(1);
        socket.setReuseAddress(true);
        socket.close();
        return port;
      } catch (final IOException e) {
        if (counter >= 20) {
          throw new IllegalStateException("Cannot open random port, tried 20 times.", e);
        }
      }
    }
  }
}
